@@ -1,4 +1,4 @@
-# Copyright (c) 2020, Riverbank Computing Limited
+# Copyright (c) 2023, Riverbank Computing Limited
 # All rights reserved.
 #
 # This copy of SIP is licensed for use under the terms of the SIP License
@@ -21,13 +21,13 @@
 # POSSIBILITY OF SUCH DAMAGE.
 
 
-from collections import OrderedDict
-from distutils.sysconfig import get_python_inc, get_python_lib
+import collections
 import os
 import packaging
 import shutil
 import subprocess
 import sys
+import sysconfig
 import tempfile
 import warnings
 
@@ -37,9 +37,8 @@ from .bindings import Bindings
 from .configurable import Configurable, Option
 from .exceptions import UserException
 from .module import resolve_abi_version
-from .py_versions import FIRST_SUPPORTED_MINOR, LAST_SUPPORTED_MINOR
-from .pyproject import (PyProjectException, PyProjectOptionException,
-        PyProjectUndefinedOptionException)
+from .py_versions import OLDEST_SUPPORTED_MINOR
+from .pyproject import PyProjectException, PyProjectOptionException
 
 
 class Project(AbstractProject, Configurable):
@@ -47,7 +46,7 @@ class Project(AbstractProject, Configurable):
 
     # The configurable options.
     _options = (
-        # The ABI version that the sip module should use.
+        # The minimum required ABI version of the sip module.
         Option('abi_version'),
 
         # The callable that will return an Bindings instance.  This is used for
@@ -66,10 +65,6 @@ class Project(AbstractProject, Configurable):
         # The list of GUI script entry points.
         Option('gui_scripts', option_type=list),
 
-        # The minimum GLIBC version required by the project.  This is used to
-        # determine the correct platform tag to use for Linux wheels.
-        Option('minimum_glibc_version'),
-
         # The minimum macOS version required by the project.  This is used to
         # determine the correct platform tag to use for macOS wheels.
         Option('minimum_macos_version'),
@@ -78,7 +73,7 @@ class Project(AbstractProject, Configurable):
         Option('py_debug', option_type=bool),
 
         # The name of the directory containing Python.h.
-        Option('py_include_dir', default=get_python_inc()),
+        Option('py_include_dir', default=sysconfig.get_path('include')),
 
         # The name of the target Python platform.
         Option('py_platform'),
@@ -105,13 +100,14 @@ class Project(AbstractProject, Configurable):
         # The fully qualified name of the sip module.
         Option('sip_module'),
 
-        # The list of files and directories, specified as glob patterns
-        # relative to the project directory, that should be included in a
-        # wheel.  If an element of list is a string then it is a pattern and
-        # files and directories are installed in the target directory.  If an
-        # element is a 2-tuple then the first part is the pattern and the
-        # second part is the name of a sub-directory relative to the target
-        # directory where the files and directories are installed.
+        # The list of files and directories, specified as glob patterns, that
+        # should be included in a wheel.  If a pattern is relative then it is
+        # taken as being relative to the project directory.  If an element of
+        # the list is a string then it is a pattern and files and directories
+        # are installed in the target directory.  If an element is a 2-tuple
+        # then the first part is the pattern and the second part is the name of
+        # a sub-directory relative to the target directory where the files and
+        # directories are installed.
         Option('wheel_includes', option_type=list),
 
         # The user-configurable options.
@@ -130,11 +126,25 @@ class Project(AbstractProject, Configurable):
                 help="disable the use of manylinux in the platform tag used "
                         "in the wheel name",
                 tools=['wheel']),
-        Option('target_dir', default=get_python_lib(plat_specific=1),
+        Option('minimum_glibc_version',
+                help="the minimum GLIBC version to be used in the platform "
+                        "tag of Linux wheels",
+                metavar="M.N", tools=['wheel']),
+        Option('scripts_dir', default=os.path.dirname(sys.executable),
+                help="the scripts installation directory", metavar="DIR",
+                tools=['build', 'install']),
+        Option('target_dir', default=sysconfig.get_path('platlib'),
                 help="the target installation directory", metavar="DIR",
                 tools=['build', 'install']),
         Option('api_dir', help="generate a QScintilla .api file in DIR",
                 metavar="DIR"),
+        Option('compile', option_type=bool, inverted=True,
+                help="disable the compilation of the generated code",
+                tools=['build']),
+        Option('version_info', option_type=bool, inverted=True,
+                help="disable any reference to the SIP version number in "
+                        "generated code",
+                tools=['build']),
     )
 
     # The configurable options for multiple bindings.
@@ -148,11 +158,12 @@ class Project(AbstractProject, Configurable):
     def __init__(self, **kwargs):
         """ Initialise the project. """
 
-        super().__init__(**kwargs)
+        super().__init__()
 
         # The current directory should contain the .toml file.
         self.root_dir = os.getcwd()
-        self.bindings = OrderedDict()
+        self.arguments = None
+        self.bindings = collections.OrderedDict()
         self.bindings_factories = []
         self.builder = None
         self.buildables = []
@@ -161,6 +172,8 @@ class Project(AbstractProject, Configurable):
         self._metadata_overrides = None
         self._temp_build_dir = None
 
+        self.initialise_options(kwargs)
+
     def apply_nonuser_defaults(self, tool):
         """ Set default values for non-user options that haven't been set yet.
         """
@@ -172,18 +185,22 @@ class Project(AbstractProject, Configurable):
             self.bindings_factory = self.import_callable(self.bindings_factory,
                     Bindings)
 
+        if self.py_major_version is None or self.py_minor_version is None:
+            self.py_major_version = sys.hexversion >> 24
+            self.py_minor_version = (sys.hexversion >> 16) & 0x0ff
+
         if self.builder_factory is None:
-            from .distutils_builder import DistutilsBuilder
-            self.builder_factory = DistutilsBuilder
+            if (self.py_major_version, self.py_minor_version) >= (3, 10):
+                from .setuptools_builder import SetuptoolsBuilder
+                self.builder_factory = SetuptoolsBuilder
+            else:
+                from .distutils_builder import DistutilsBuilder
+                self.builder_factory = DistutilsBuilder
         elif isinstance(self.builder_factory, str):
             # Convert the name to a callable.
             self.builder_factory = self.import_callable(self.builder_factory,
                     AbstractBuilder)
 
-        if self.py_major_version is None or self.py_minor_version is None:
-            self.py_major_version = sys.hexversion >> 24
-            self.py_minor_version = (sys.hexversion >> 16) & 0x0ff
-
         if self.py_platform is None:
             self.py_platform = sys.platform
 
@@ -195,11 +212,6 @@ class Project(AbstractProject, Configurable):
     def apply_user_defaults(self, tool):
         """ Set default values for user options that haven't been set yet. """
 
-        # If we are the backend to a 3rd-party frontend (most probably pip)
-        # then let it handle the verbosity of messages.
-        if self.verbose is None and tool == '':
-            self.verbose = True
-
         # This is only used when creating sdist and wheel files.
         if self.name is None:
             self.name = self.metadata['name']
@@ -306,6 +318,46 @@ class Project(AbstractProject, Configurable):
 
         return ''
 
+    def get_platform_tag(self):
+        """ Return the platform tag to use in a wheel name.  This default
+        implementation uses the platform name and applies PEP defined
+        conventions depending on OS version and GLIBC version as appropriate.
+        """
+
+        platform_tag = sysconfig.get_platform()
+
+        if self.py_platform == 'darwin' and self.minimum_macos_version:
+            # We expect a three part tag so leave anything else unchanged.
+            parts = platform_tag.split('-')
+            if len(parts) == 3:
+                min_major = int(self.minimum_macos_version[0])
+                min_minor = int(self.minimum_macos_version[1])
+
+                # For arm64 binaries enforce a valid minimum macOS version.
+                if parts[2] == 'arm64' and min_major < 11:
+                    min_major = 11
+                    min_minor = 0
+
+                parts[1] = '{}.{}'.format(min_major, min_minor)
+
+                platform_tag = '-'.join(parts)
+
+        elif self.py_platform == 'linux' and self.manylinux:
+            # We expect a two part tag so leave anything else unchanged.
+            parts = platform_tag.split('-')
+            if len(parts) == 2:
+                if self.minimum_glibc_version:
+                    major, minor = self.minimum_glibc_version
+                else:
+                    major, minor = 2, 5
+
+                parts[0] = 'manylinux'
+                parts.insert(1, '{}.{}'.format(major, minor))
+
+                platform_tag = '-'.join(parts)
+
+        return platform_tag.replace('.', '_').replace('-', '_')
+
     def get_requires_dists(self):
         """ Return any 'Requires-Dist' to add to the project's meta-data. """
 
@@ -332,7 +384,7 @@ class Project(AbstractProject, Configurable):
                 next_abi_major)]
 
     def get_sip_distinfo_command_line(self, sip_distinfo, inventory,
-            generator=None, wheel_tag=None):
+            generator=None, wheel_tag=None, generator_version=None):
         """ Return a sequence of command line arguments to invoke sip-distinfo.
         """
 
@@ -353,6 +405,10 @@ class Project(AbstractProject, Configurable):
             args.append('--generator')
             args.append(generator)
 
+        if generator_version is not None:
+            args.append('--generator-version')
+            args.append(generator_version)
+
         if wheel_tag is not None:
             args.append('--wheel-tag')
             args.append(wheel_tag)
@@ -387,6 +443,62 @@ class Project(AbstractProject, Configurable):
         self.builder.install()
         self._remove_build_dir()
 
+    @property
+    def minimum_glibc_version(self):
+        """ The getter for the minimum GLIBC version. """
+
+        return self._minimum_glibc_version
+
+    @minimum_glibc_version.setter
+    def minimum_glibc_version(self, value):
+        """ The setter for the minimum GLIBC version. """
+
+        # Handle the initial creation of the option value.
+        if value is None and not hasattr(self, '_minimum_glibc_version'):
+            self._minimum_glibc_version = None
+            return
+
+        # Make sure any minimum GLIBC version is valid and convert it to a
+        # 2-tuple.
+        if value:
+            try:
+                value = self._convert_major_minor(value)
+            except ValueError:
+                raise PyProjectOptionException('minimum-glibc-version',
+                        "'{0}' is an invalid GLIBC version number".format(
+                                value),
+                        section_name='tool.sip.project')
+
+        self._minimum_glibc_version = value
+
+    @property
+    def minimum_macos_version(self):
+        """ The getter for the minimum macOS version. """
+
+        return self._minimum_macos_version
+
+    @minimum_macos_version.setter
+    def minimum_macos_version(self, value):
+        """ The setter for the minimum macOS version. """
+
+        # Handle the initial creation of the option value.
+        if value is None and not hasattr(self, '_minimum_macos_version'):
+            self._minimum_macos_version = None
+            return
+
+        # Make sure any minimum macOS version is valid and convert it to a
+        # 2-tuple.
+        if value:
+            try:
+                value = self._convert_major_minor(value)
+            except ValueError:
+                raise PyProjectOptionException('minimum-macos-version',
+                        "'{0}' is an invalid macOS version number".format(
+                                value),
+                        section_name='tool.sip.project')
+
+        self._minimum_macos_version = value
+
     @staticmethod
     def open_for_writing(fname):
         """ Open a file for writing while handling any errors. """
@@ -461,14 +573,9 @@ class Project(AbstractProject, Configurable):
         # Set the initial configuration from the pyproject.toml file.
         self._set_initial_configuration(pyproject, tool)
 
-        # Add any tool-specific command line options for (so far unspecified)
+        # Add any tool-specific command line arguments for (so far unspecified)
         # parts of the configuration.
-        if tool != 'pep517':
-            self._configure_from_command_line(tool, tool_description)
-        else:
-            # Until pip improves it's error reporting we give the user all the
-            # help we can.
-            self.verbose = True
+        self._configure_from_arguments(tool, tool_description)
 
         # Now that any help has been given we can report a problematic
         # pyproject.toml file.
@@ -528,6 +635,9 @@ class Project(AbstractProject, Configurable):
             if not b.is_buildable():
                 del self.bindings[b.name]
 
+        if len(self.bindings) == 0:
+            raise UserException("There are no bindings that can be built")
+
     def verify_configuration(self, tool):
         """ Verify that the configuration is complete and consistent. """
 
@@ -537,49 +647,25 @@ class Project(AbstractProject, Configurable):
                     "'{0}' must begin with a digit".format(self.build_tag),
                     section_name='tool.sip.project')
 
-        # Make sure any minimum GLIBC version is valid and convert it to a
-        # 2-tuple.
-        if self.minimum_glibc_version:
-            try:
-                self.minimum_glibc_version = self._convert_major_minor(
-                        self.minimum_glibc_version)
-            except ValueError:
-                raise PyProjectOptionException('minimum-glibc-version',
-                        "'{0}' is an invalid GLIBC version number".format(
-                                self.minimum_glibc_version),
-                        section_name='tool.sip.project')
-        else:
-            self.minimum_glibc_version = (2, 5)
-
-        # Make sure any minimum macOS version is valid and convert it to a
-        # 2-tuple.
-        if self.minimum_macos_version:
-            try:
-                self.minimum_macos_version = self._convert_major_minor(
-                        self.minimum_macos_version)
-            except ValueError:
-                raise PyProjectOptionException('minimum-macos-version',
-                        "'{0}' is an invalid macOS version number".format(
-                                self.minimum_macos_version),
-                        section_name='tool.sip.project')
-
         # Make sure relevent paths are absolute and use native separators.
         self.sip_files_dir = self.project_path(self.sip_files_dir)
         self.sip_include_dirs = [self.project_path(d)
                 for d in self.sip_include_dirs]
 
-        # Make sure we support the targeted version of Python.
+        # Check that the targeted version of Python isn't too old.  We hope
+        # that we will support newer versions automatically, but it's not
+        # guaranteed.
         py_version = (self.py_major_version, self.py_minor_version)
-        first_version = (3, FIRST_SUPPORTED_MINOR)
-        last_version = (3, LAST_SUPPORTED_MINOR)
+        first_version = (3, OLDEST_SUPPORTED_MINOR)
 
-        if py_version < first_version or py_version > last_version:
+        if py_version < first_version or self.py_major_version > 3:
             raise UserException(
                     "Python v{}.{} is not supported".format(
                             self.py_major_version, self.py_minor_version))
 
-        # Make sure we have a valid ABI version.
-        self.abi_version = resolve_abi_version(self.abi_version)
+        # Get the supported ABI version.  (The actual version may have a later
+        # minor version.)
+        self.abi_version = resolve_abi_version(self.abi_version, module=False)
 
         # Checks for standalone projects.
         if tool in Option.BUILD_TOOLS and not self.sip_module:
@@ -625,8 +711,8 @@ class Project(AbstractProject, Configurable):
         for bindings in self.bindings.values():
             bindings.verify_configuration(tool)
 
-    def _configure_from_command_line(self, tool, tool_description):
-        """ Update the configuration from the user supplied command line. """
+    def _configure_from_arguments(self, tool, tool_description):
+        """ Update the configuration from any user supplied arguments. """
 
         from argparse import SUPPRESS
         from .argument_parser import ArgumentParser
@@ -652,7 +738,7 @@ class Project(AbstractProject, Configurable):
             bindings.add_command_line_options(parser, tool, all_options)
 
         # Parse the arguments and update the corresponding configurables.
-        args = parser.parse_args()
+        args = parser.parse_args(self.arguments)
 
         for option, configurables in all_options.items():
             for configurable in configurables:
